#!/usr/bin/perl

#Version 1.00

use strict;

use HTTP::Proxy;
use HTTP::Proxy::BodyFilter::complete;
use HTML::TreeBuilder;

# initialisation
my $proxy = HTTP::Proxy->new( port => 3128 );
close STDERR;

# This logs the queries and clicks
$proxy->push_filter( request => LogClicks->new() );

# By forcing the client not to cache, we get all clicks, not just first-time clicks
# Depending on the setting, we might not want to do this.
$proxy->push_filter( response => NoCache->new() );

# Changes the user agent to avoid content from Bing being compressed
$proxy->push_filter( host => 'bing.com',  request => NoCompression->new());

# This catches the results, rewriting the content as well (only on Bing requests)
$proxy->push_filter( host => 'bing.com', 
					 response => HTTP::Proxy::BodyFilter::complete->new(), 
					 response => RewriteContent->new() );

$proxy->start;

############### CLICK LOGGING ################
{
    package LogClicks;
    use base qw( HTTP::Proxy::HeaderFilter );

    sub filter {
        my ( $self, $headers, $message ) = @_;

        my $url = $message->url;
		my $referer = $headers->header('Referer');
		
		# If the referer is Bing, its a follow on request
		# Only keep non-bing requests
		if ($referer =~ /bing.com\/search\?q=([^&]*)/ && $url !~ /bing.com/ && $url !~ /bing.net/) 
		{
			my $query = $1;
			my $date = scalar localtime;
			my $ip = $headers->header('X-Forwarded-For');
						
			print "CLICK\t$date\t$query\t$ip\t$url\n";
		}
    }
}

########### MODIFY THE URL ###############################
#{
#	package ThirtyResults;
#	use base qw ( HTTP::Proxy::BodyFilter );
#	
#	sub filter {
#		my ( $self, $dataref, $request, $protocol, $buffer ) = @_;
#		
#		my $uri = $request->uri;
#		# For interesting reranking, we want more results. So we modify the URL
#		if ($uri =~ /search\?q=/ && $uri !~ /count=/) {
#			$request->uri("$uri&count=30");
#		}
#	}
#}

########### RESULT LOGGING, CONTENT REWRITING ############
{
    package RewriteContent;
    use base qw( HTTP::Proxy::BodyFilter );

    sub filter {
        my ( $self, $dataref, $response, $protocol, $buffer ) = @_;

        my $url = $response->base;

        # If its a Bing request, and its the last chunk of the request (all the data is present)
        if ($url =~ /^http:\/\/www.bing.com\/search?/ && !defined($buffer)) 
		{
			# Rerank and interleave results
			# 1. Cut out results, and large graphical elements (sb_ans)
			my $tree = HTML::TreeBuilder->new;
			$tree->parse_content($$dataref);
			my @results;
			foreach my $r ($tree->look_down('_tag', 'li')) 
			{
				# Standard URLs
				if ($r && $r->attr('class') eq "sa_wr") {
					$results[$#results+1] = $r->as_HTML();
				}
				# Large graphical elements
				if ($r && $r->attr('class') eq "sb_ans") {
					$results[$#results+1] = $r->as_HTML();
				}
			}
			
			# 2. Make a reranked result set
			my @reranked = (sort simpleRanker @results);

			# 3. Interleave
			my $teams;
			my $interleavedRef = interleave(\@results, \@reranked, \$teams);
			
			#4. Paste in a new top 10
			my $newResults = join("</li>", (@$interleavedRef)) . "</li>";
			$$dataref =~ s|(<ul[^>]*sb_results[^>]*>)<li.*?</div></li></ul>|${1}RESULTSHERE</ul>|s;
			substr($$dataref, index($$dataref, "RESULTSHERE"), length("RESULTSHERE"), $newResults);
						
			# 5. Log the query
			if ($url =~ /\?q=([^&]*)/) 
			{
				my $query = $1;
				my $date = scalar localtime;
				my $ip = $response->request->header('X-Forwarded-For');
								
				# Collect the results
				my $results = "";
				while ($$dataref =~ /<h3><a[^>]*href="(http[^"]*)"[^>]*SERP/g)
				{
					$results .= "$1;";
				}

				my $orig = record(\@results);
				my $new = record(\@reranked);
				print "QUERY\t$date\t$query\t$ip\t$teams\t$results\t$orig\t$new\n";
			}
        }
    }
	
	sub interleave
	{
		my ($rankingAref, $rankingBref, $teamRef) = (shift, shift, shift);
		my @interleaved;

		# Toss a coin 5 times, to make a list of 10 results
		for (my $i=0; $i<5; $i++) {

			my $coin = rand>0.5?"A":"B";
		
			if ($coin eq "A") { # Ranker A won the coin toss
				my $nextA = computeNext($rankingAref, \@interleaved);
				push @interleaved, $rankingAref->[$nextA];
				my $nextB = computeNext($rankingBref, \@interleaved);
				push @interleaved, $rankingBref->[$nextB];
				$$teamRef .= "AB";

			} else { # Ranker B won the coin toss
				my $nextB = computeNext($rankingBref, \@interleaved);
				push @interleaved, $rankingBref->[$nextB];
				my $nextA = computeNext($rankingAref, \@interleaved);
				push @interleaved, $rankingAref->[$nextA];			
				$$teamRef .= "BA";
			}
		}
				
		# Check for shared results, set back to neither team
		for (my $i=0; $i<=$#interleaved; $i++) {
			if ($interleaved[$i] eq $rankingAref->[$i] && $interleaved[$i] eq $rankingBref->[$i]) {
				substr ($$teamRef, $i, 1, "s");
			} else {
				last;
			}
		}
		
		# For the tutorial - add team annotation
      for (my $i=0; $i<=$#interleaved; $i++) {
      	my $team = substr($$teamRef, $i, 1);
      	if ($team eq "A") { $interleaved[$i] =~ s|"sb_tlst">|"sb_tlst"><div style=\"background:#ff6666\">[Orig]</div> |; }
      	if ($team eq "B") { $interleaved[$i] =~ s|"sb_tlst">|"sb_tlst"><div style=\"background:#66dd11\">[New]</div> |; }
      	if ($team eq "s") { $interleaved[$i] =~ s|"sb_tlst">|"sb_tlst"><div style=\"background:#7777ff\">[shared]</div> |; }
      }
			
		return \@interleaved;
	}
	
	sub computeNext
	{
		my ($rankingRef, $takenRef) = (shift, shift);
		my $next = 0;
		# While the result $rankingRef->[$next] is in the taken list, keep looking
		while ( $next <= $#$rankingRef ) 
		{
			my $url = $rankingRef->[$next];
			my $found = 0;
			foreach my $taken (@$takenRef) 
			{
				if ($url eq $taken) {
					$found = 1;
				}
			}
			if ($found == 0)
			{
				return $next;
			}
			$next++;
		}
		
		return $next;
	}
	
	# Make a list of URLs
	sub record
	{
		my $resultsRef = shift;
		my $results = join " ", @$resultsRef;

		my $output;
		while ($results =~ /<h3><a[^>]*href="([^"]*)"[^>]*SERP/g)
		{
			$output .= "$1;";
		}
		
		return $output;
	}
	
	##################### SIMPLE RERANKER #######################
	
	# Rank by how many of our favorite or least-favorite words occur in search results
	sub simpleRanker
	{
		my @goodWords = ("acm", "dblp", "edu", "2011");
		my @badWords = ("wikipedia", "answers");
		
		# $a and $b are the two input URLs. 
		my $scoreA = countWords($a, \@goodWords) - countWords($a, \@badWords);
		my $scoreB = countWords($b, \@goodWords) - countWords($b, \@badWords);
		
		# Returns +1 if URL $b scores higher, -1 if $a scores higher
		return ($scoreB <=> $scoreA);
	}
	
	# Count how many times each word occurs
	sub countWords
	{
		my ($text, $wordsRef) = (shift, shift);
		my $n = 0;
		
		foreach my $word (@$wordsRef) 
		{
			while ($text =~ /$word/ig)
			{
				$n++;
			}
		}
		return $n;
	}

}

############## HEADER REWRITING ##############
{
    package NoCompression;
    use base qw( HTTP::Proxy::HeaderFilter );

    # changes the User-Agent header in all requests
    sub filter {
        my ( $self, $headers, $message ) = @_;

		$message->headers->header( User_Agent => 'Mozilla/4.0 (compatible) Perl proxy' );
		$message->headers->header( Accept => 'application/xml,text/xml' );
    }
}

{
	package NoCache;
    use base qw( HTTP::Proxy::HeaderFilter );

	# Tells the web browser not to cache pages where the referer is Bing
    sub filter {
        my ( $self, $headers, $message ) = @_;

		my $referer = $message->request->headers->header("referer");
		if ($referer =~ /bing/) {
			$message->headers->header( 'Cache-Control' => 'no-cache' );
		}
    }
}
	
